/*
 * L9CutForm.cs - Dialog frontend for the L9cut utility.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.IO;
using System.Drawing;
using System.Windows.Forms;

namespace Level9
{
	/// <summary>
	/// Dialog frontend for the L9cut utility.
	/// </summary>
	public partial class L9CutForm : Form
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The default L9Cut output file.
		/// </summary>
		const string DEFAULT_OUTPUT_FILE = "\\output.dat";

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// This property indicates whether the selected arguments are valid
		/// for an execution of L9cut.
		/// </summary>
		bool checkExecute = true;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the L9CutForm class.
		/// </summary>
		public L9CutForm()
		{
			InitializeComponent();
			//
			// Set button and label images
			//
			buttonExecute.Image = Images.Get(Images.Stock.Execute_16x16);
			buttonClose.Image   = Images.Get(Images.Stock.Close_16x16);
			labelL9Cut.Image    = Images.Get(Images.Stock.L9cut_48x48);
		}

		// -- Event handling -------------------------------------------------

		/// <summary>
		/// Invoked when the input file changes.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void TextBoxInputTextChanged(object sender, EventArgs e)
		{
			if (checkExecute) {
				CheckExecuteExtract();
			}
		}

		/// <summary>
		/// Invoked when the output file changes.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void TextBoxOutputTextChanged(object sender, EventArgs e)
		{
			if (checkExecute) {
				CheckExecuteExtract();
			}
		}

		/// <summary>
		/// Invoked when the 'Detected datafiles' button changes its state.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void RadioButtonDetectedDatafilesCheckedChanged(object sender,
			EventArgs e)
		{
			if (checkExecute) {
				CheckExecuteLists();
			}
		}

		/// <summary>
		/// Invoked when the 'Missing datafiles' button changes its state.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void RadioButtonMissingDatafilesCheckedChanged(object sender,
			EventArgs e)
		{
			if (checkExecute) {
				CheckExecuteLists();
			}
		}

		/// <summary>
		/// Invoked when the 'Missing patches' button changes its state.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void RadioButtonMissingPatchesCheckedChanged(object sender,
			EventArgs e)
		{
			if (checkExecute) {
				CheckExecuteLists();
			}
		}

		/// <summary>
		/// Invoked when the 'Execute' button has been pressed.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void ButtonExecuteClick(object sender, EventArgs e)
		{
			if (radioButtonDetectedDatafiles.Checked) {
				richTextBoxOutput.Text = L9Cut.GetDetectedDatafiles();
			}
			else if (radioButtonMissingDatafiles.Checked) {
				richTextBoxOutput.Text = L9Cut.GetMissingDatafiles();
			}
			else if (radioButtonMissingPatches.Checked) {
				richTextBoxOutput.Text = L9Cut.GetMissingPatches();
			}
			else {
				richTextBoxOutput.Text = L9Cut.ExtractDatafile(
					textBoxInput.Text, textBoxOutput.Text,
					checkBoxProtection.Checked ? true : false);
			}
			if (richTextBoxOutput.Text != null) {
				richTextBoxOutput.SelectionStart = 0;
				richTextBoxOutput.ScrollToCaret();
			}
		}

		/// <summary>
		/// Invoked when the 'Close' button has been pressed.
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void ButtonCloseClick(object sender, EventArgs e)
		{
			this.DialogResult = DialogResult.OK;
			this.Close();
		}

		/// <summary>
		/// Invoked when the 'Browse' button has been pressed (input file).
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">an EventArgs that contains the event data</param>
		void ButtonBrowseInputClick(object sender, EventArgs e)
		{
			string filename = FileUtils.OFDDataFiles(labelInputFile.Text);

			if (filename == null) {
				return;
			}

			textBoxInput.Text = filename;
			textBoxInput.SelectionStart = textBoxInput.TextLength - 1;
			textBoxOutput.Text = Path.GetDirectoryName(filename)
				+ DEFAULT_OUTPUT_FILE;
			textBoxOutput.SelectionStart = textBoxOutput.TextLength - 1;
		}

		/// <summary>
		/// Invoked when the 'Browse' button has been pressed (output file).
		/// </summary>
		/// <param name="sender">the object that originated the event</param>
		/// <param name="e">event arguments</param>
		void ButtonBrowseOutputClick(object sender, EventArgs e)
		{
			string filename = FileUtils.SFDDataFiles(labelOutputFile.Text);

			if (filename == null) {
				return;
			}

			textBoxOutput.Text = filename;
			textBoxOutput.SelectionStart = textBoxOutput.TextLength - 1;
		}

		// -- Methods --------------------------------------------------------

		/// <summary>
		/// Checks the selected arguments if the detected datafiles or the
		/// missing datafiles or the missing patches are to be displayed.
		/// </summary>
		void CheckExecuteLists()
		{
			checkExecute = false;
			if (radioButtonDetectedDatafiles.Checked
					|| radioButtonMissingDatafiles.Checked
					|| radioButtonMissingPatches.Checked) {
				textBoxInput.Clear();
				textBoxOutput.Clear();
				checkBoxProtection.Checked = false;
				buttonExecute.Enabled = true;
			}
			else {
				buttonExecute.Enabled = false;
			}
			checkExecute = true;
		}

		/// <summary>
		/// Checks the selected arguments if a datafile is to be extracted.
		/// </summary>
		void CheckExecuteExtract()
		{
			checkExecute = false;
			radioButtonDetectedDatafiles.Checked = false;
			radioButtonMissingDatafiles.Checked  = false;
			radioButtonMissingPatches.Checked    = false;
			if (textBoxInput.TextLength > 0
					&& textBoxOutput.TextLength > 0) {
				buttonExecute.Enabled = true;
			}
			else {
				buttonExecute.Enabled = false;
			}
			checkExecute = true;
		}
	}
}
